/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mnenhy.
 *
 * The Initial Developer of the Original Code is Karsten Düsterloh <mnenhy@tprac.de>.
 *
 * Portions created by the Initial Developer are Copyright (C) 2002-2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// training.dat format
//
//    [0xFEEDFACE]
//    [number good messages][number bad messages]
//    [number good tokens]
//    [refcount][length of token]token
//    ....
//    [number bad tokens]
//    [refcount][length of token]token
//


// |
// |  constants
// |
const gksTraining           = "training.dat";
const gksXULTable           = "mnenhyJunkTable";
const gksXULProlog          = "mnenhyJunkProlog";
const gksXULGoodProgress    = "mnenhyGoodProgress";
const gksXULEvilProgress    = "mnenhyEvilProgress";
const gkoProperties         = goMnenhy.InitStringBundle("chrome://mnenhy/locale/mnenhy-junk.properties")
const gksJunkProlog         = goMnenhy.GetString("Prolog",   gkoProperties);
const gksJunkNotFound       = goMnenhy.GetString("NotFound", gkoProperties);
const gksJunkInvalid        = goMnenhy.GetString("Invalid",  gkoProperties);
// start new reading thread every |gkiChunkSize| tokens
const gkiChunkSize          = goMnenhy.GetPref("extensions.mnenhy.junk.chunk", goMnenhy.kiPrefInt) || 250;
// start new table row thread every |gkiTableChunkSize| tokens
const gkiTableChunkSize     = goMnenhy.GetPref("extensions.mnenhy.junk.table", goMnenhy.kiPrefInt) ||  30;
// ms delay for calls to setTimeout
const gkiTimeoutDelay       = goMnenhy.GetPref("extensions.mnenhy.junk.delay", goMnenhy.kiPrefInt);
// UTF-8 handling
const gkoUTFConv            = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"]
                                        .createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
gkoUTFConv.charset = "UTF-8";
// UTF-8 to HTML
const gkoTXTToHTMLConv      = Components.classes["@mozilla.org/txttohtmlconv;1"]
                                        .createInstance(Components.interfaces.mozITXTToHTMLConv);


// |
// |  variables
// |
var giMinRefCount = goMnenhy.GetPref("extensions.mnenhy.junk.threshold", goMnenhy.kiPrefInt);
var goTrainingDat =
{ // training.dat
  moFile:       null,     // the training.dat file
  moIS:         null,     // input stream
  moData:       {},       // the training.dat data
  // statistical data
  miGoodMsgs:   0,        // number of good messages
  miEvilMsgs:   0,        // number of evil messages
  miGoodTokens: 0,        // number of good tokens
  miEvilTokens: 0,        // number of evil tokens
  miGoodEvil:   0,        // number of tokens both good and evil
  miSkipped:    0,        // number of tokens below threshold
  miShown:      0,        // number of tokens shown in table
  miStartTime:  0,        // start time reading
  miEndTime:    0,        // end time reading
  // helpers
  moProgress:   null      // DOM element for progress display
};
var goSorter = null;      // sort class for table column sort


// |
// |  helper functions
// |

// get the training.dat
function GetJunkStatFile()
{
  var sBaseDir = goMnenhy.m_koDirService.get("ProfD", Components.interfaces.nsIFile);
  var oFile    = new goMnenhy.CFileByFile(sBaseDir); // dir
  oFile.append(gksTraining);                         // + name
  return oFile;
}


// create XHTML-DOM nodes from string and attach them to oParent
function AddInnerXML(oParent, sXML)
{ // set new
  var oDOMParser = new DOMParser();
  var oNodes     = oDOMParser.parseFromString("<dummy xmlns='http://www.w3.org/1999/xhtml'>"+sXML+"</dummy>", "text/xml").firstChild.childNodes;
  if (!oNodes || (oNodes.length && oNodes[0].nodeName=="#text") )
    goMnenhy.Log("AddInnerXML="+sXML) 
  for (var i = 0; i < oNodes.length; ++i)
    oParent.appendChild(oNodes[i]);
}


// create XHTML-DOM nodes from string and attach only them to oNode
function SetInnerXML(oNode, sXML)
{ // delete child nodes first
  while (oNode.hasChildNodes())
    oNode.removeChild(oNode.lastChild);
  // set new
  AddInnerXML(oNode, sXML);
}


// compute the modified bayesian junk probality of a word
// (see  nsBayesianFilter::classifyMessage in
//  /seamonkey/source/mailnews/extensions/bayesian-spam-filter/src/nsBayesianFilter.cpp)
// return a percent value
function GetJunkProb(aiGood, aiEvil)
{ // aiGood    = good occurences of this word
  // aiEvil    = evil occurences of this word
  const kiGoodMsg = goTrainingDat.miGoodMsgs || 1; // avoid NaN
  const kiEvilMsg = goTrainingDat.miEvilMsgs || 1; // avoid NaN

  var prob = (aiEvil / kiEvilMsg) / (aiGood / kiGoodMsg + aiEvil / kiEvilMsg);
  var n    = aiEvil + aiGood;
  prob = (.225 + n * prob) / (.45 + n);
  return 100 * prob;
}


// |
// |  main functions
// |

// get the training.dat and prepare reading
function StartReading()
{
  goMnenhy.SetWaitCursor(window);
  // get training.dat
  goTrainingDat.moFile = GetJunkStatFile();
  // check junk threshold (0 = automatic threshold)
  if (!giMinRefCount)
  { // round kiSize
    const kiSize = goTrainingDat.moFile.exists() ? (goTrainingDat.moFile.fileSize / 1000) : 0;
    giMinRefCount = parseInt(kiSize, 10);
  }
  // open training.dat
  goTrainingDat.moIS = goMnenhy.GetBinStream(goTrainingDat.moFile);
  if (!goTrainingDat.moIS)
    SetInnerXML(document.getElementById(gksXULProlog), gksJunkNotFound);
  else
  { // start time counter
    goTrainingDat.miStartTime = new Date().getTime() / 1000;

    // check magic number
    var iMagicNumber = goTrainingDat.moIS.read32();
    if (iMagicNumber != 0xFEEDFACE)
      SetInnerXML(document.getElementById(gksXULProlog), gksJunkInvalid);
    else
    { // set ham'n'spam numbers
      goTrainingDat.miGoodMsgs = goTrainingDat.moIS.read32();
      goTrainingDat.miEvilMsgs = goTrainingDat.moIS.read32();
      // start reading good tokens
      goTrainingDat.moProgress   = document.getElementById(gksXULGoodProgress);
      goTrainingDat.miGoodTokens = goTrainingDat.moIS.read32();
      setTimeout(ReadGoodTokens, gkiTimeoutDelay, 0);
    } // magic number found
  } // have stream
}


// Read good token data
// IN: int aiRead: number of good tokens already read
// To avoid locking the UI, we call ourselves every |gkiChunkSize| tokens!
function ReadGoodTokens(aiRead)
{
  const kiMax = Math.min(goTrainingDat.miGoodTokens, aiRead + gkiChunkSize);
  var oToken, iRefCount, iTokenLen, sToken;

  for (; aiRead < kiMax; ++aiRead)
  { // read token data (add space to avoid problems with keywords)
    iRefCount  = goTrainingDat.moIS.read32();
    iTokenLen  = goTrainingDat.moIS.read32();
    sToken     = goTrainingDat.moIS.readBytes(iTokenLen)+" ";
    // most tokens are in UTF-8 (but not all, so ConvertToUnicode may vomit :( )
    try{sToken = gkoUTFConv.ConvertToUnicode(sToken)}catch(ignore){}
    // good elements are always new
    goTrainingDat.moData[sToken] = {iGood: iRefCount, iEvil: 0};
  }

  if (goTrainingDat.moProgress)
    goTrainingDat.moProgress.textContent = aiRead + "\xA0/\xA0" + goTrainingDat.miGoodTokens;

  if (aiRead < goTrainingDat.miGoodTokens)
  { // if we have good tokens left to read, do it
    setTimeout(ReadGoodTokens, gkiTimeoutDelay, aiRead);
  }
  else
  { // we have no further good tokens, so prepare reading evil tokens
    goTrainingDat.moProgress   = document.getElementById(gksXULEvilProgress);
    goTrainingDat.miEvilTokens = goTrainingDat.moIS.read32();
    setTimeout(ReadEvilTokens, gkiTimeoutDelay, 0);
  }
}


// Read evil token data
// IN: int aiRead: number of evil tokens already read
// We do some optimizing here: certain tokens can already be skipped...
// (Most of this function is similar to ReadGoodTokens above,
//  but I'm trying to gain some performance by avoiding condition testing.)
// To avoid locking the UI, we call ourselves every |gkiChunkSize| tokens!
function ReadEvilTokens(aiRead)
{
  const kiMax = Math.min(goTrainingDat.miEvilTokens, aiRead + gkiChunkSize);
  var oToken, iRefCount, iTokenLen, sToken;

  for (; aiRead < kiMax; ++aiRead)
  { // read token data (add space to avoid problems with keywords)
    iRefCount  = goTrainingDat.moIS.read32();
    iTokenLen  = goTrainingDat.moIS.read32();
    sToken     = goTrainingDat.moIS.readBytes(iTokenLen)+" ";
    // most tokens are in UTF-8 (but not all, so ConvertToUnicode may vomit :( )
    try{sToken = gkoUTFConv.ConvertToUnicode(sToken)}catch(ignore){}
    // don't add known words
    if (sToken in goTrainingDat.moData)
    {
      ++goTrainingDat.miGoodEvil;
      oToken = goTrainingDat.moData[sToken]
      if (iRefCount + oToken.iGood >= giMinRefCount)
      { // add evil value to token
        oToken.iEvil = iRefCount;
      }
      else
      { // delete token
        ++goTrainingDat.miSkipped;
        delete goTrainingDat.moData[sToken];
      }
    }
    else if (iRefCount >= giMinRefCount)
    { // add token
      goTrainingDat.moData[sToken] = {iGood: 0, iEvil: iRefCount};
    }
    else
    { // don't add token
      ++goTrainingDat.miSkipped;
    }
  }

  if (goTrainingDat.moProgress)
    goTrainingDat.moProgress.textContent = aiRead + "\xA0/\xA0" + goTrainingDat.miEvilTokens;

  if (aiRead < goTrainingDat.miEvilTokens)
  { // if we have evil tokens left to read, do it
    setTimeout(ReadEvilTokens, gkiTimeoutDelay, aiRead);
  }
  else
  { // we have no further tokens
    setTimeout(FinishReading, gkiTimeoutDelay, 0);
  }
}


// all tokens read
function FinishReading()
{ // prepare HTML table creation
  var oNode = goTrainingDat.moProgress = document.getElementById(gksXULTable);
  // delete child nodes first
  while (oNode.hasChildNodes())
    oNode.removeChild(oNode.lastChild);
  setTimeout(BuildJunkTable, gkiTimeoutDelay, 0);
}


// Process the training.dat data and create HTML tabele
// We drop some tokens below the threshold
// To avoid locking the UI, we call ourselves every |gkiTableChunkSize| tokens!
// IN: int aiBuilt: number of table rows already created
function BuildJunkTable(aiBuilt)
{
  var i = 0, bChunked = false, oToken = null, sRow = "", sh, ihg, ihe, ihp;
  for (var s in goTrainingDat.moData)
  {
    ++i;
    oToken = goTrainingDat.moData[s];
    if (!oToken.iEvil && oToken.iGood < giMinRefCount)
    { // skip good-only tokens below threshold
      ++goTrainingDat.miSkipped;
    }
    else
    { // add table row
      ++aiBuilt;
      sh  = gkoTXTToHTMLConv.scanTXT(s, gkoTXTToHTMLConv.kEntities).replace(/\"/g, '&#34;');
      ihg = oToken.iGood;
      ihe = oToken.iEvil;
      ihp = GetJunkProb(oToken.iGood, oToken.iEvil).toFixed(2);
      sRow += '<tr index="'+aiBuilt+'" token="'+sh+'" good="'+ihg+'" evil="'+ihe+'" prob="'+ihp+'">'
            + "<td class='JunkCol0'>"  + String(aiBuilt)
            + "</td><td class='JunkCol1'>" + sh
            + "</td><td>"                  + ihg
            + "</td><td>"                  + ihe
            + "</td><td>"                  + ihp
            + " %</td></tr>\n";
    }
    // remove token from list
    delete goTrainingDat.moData[s];
    bChunked = i >= gkiTableChunkSize;
    if (bChunked)
      break;
  }
  if (sRow)
    AddInnerXML(goTrainingDat.moProgress, sRow);

  if (bChunked)
  { // if we have tokens left, loop
    setTimeout(BuildJunkTable, gkiTimeoutDelay, aiBuilt);
  }
  else
  { // we have no further tokens
    setTimeout(BuildJunkProlog, gkiTimeoutDelay);
  }
}


function BuildJunkProlog()
{ // stop timer
  goTrainingDat.miEndTime = new Date().getTime() / 1000;
  // percentages of messages
  const kiMsgs          = goTrainingDat.miGoodMsgs + goTrainingDat.miEvilMsgs;
  const kiGoodP         = Math.round(goTrainingDat.miGoodMsgs / kiMsgs * 100);
  const kiEvilP         = 100 - kiGoodP;
  // duration of reading the training.dat
  const kiDuration      = parseInt(100 * (goTrainingDat.miEndTime - goTrainingDat.miStartTime), 10) / 100;
  // percentages of tokens
  const kiTokens        = goTrainingDat.miGoodTokens + goTrainingDat.miEvilTokens;
  const kiGoodTP        = parseInt(goTrainingDat.miGoodTokens / kiTokens * 100, 10);
  const kiEvilTP        = 100 - kiGoodTP;
  // number of different and ignored tokens
  const kiDifferent     = kiTokens    - goTrainingDat.miGoodEvil;
  goTrainingDat.miShown = kiDifferent - goTrainingDat.miSkipped;
  const kiSize          = goTrainingDat.moFile.exists() ? goTrainingDat.moFile.fileSize : 0;

  var sProlog = gksJunkProlog.replace("%m",  String(kiMsgs))
                             .replace("%h",  String(goTrainingDat.miGoodMsgs))
                             .replace("%hp", String(kiGoodP))
                             .replace("%s",  String(goTrainingDat.miEvilMsgs))
                             .replace("%sp", String(kiEvilP))
                             .replace("%b",  String(kiSize))
                             .replace("%d",  String(kiDuration))
                             .replace("%j",  String(kiTokens))
                             .replace("%g",  String(goTrainingDat.miGoodTokens))
                             .replace("%gp", String(kiGoodTP))
                             .replace("%e",  String(goTrainingDat.miEvilTokens))
                             .replace("%ep", String(kiEvilTP))
                             .replace("%ge", String(kiDifferent))
                             .replace("%mi", String(goTrainingDat.miShown))
                             .replace("%i",  String(goTrainingDat.miSkipped))
                             .replace("%t",  String(giMinRefCount));
  SetInnerXML(document.getElementById(gksXULProlog), sProlog);
  
  // prepare the sorter
  eval("goSorter=" + goMnenhy.CSort.toSource());
  goSorter = new goSorter(goTrainingDat.moProgress, "", false, true, window);

  goMnenhy.UnsetWaitCursor(window);
}


function SortJunkTable(asSortAttr, abAscending, abNumber)
{
  goMnenhy.SetWaitCursor(window);

  goSorter.msAttr = asSortAttr;
  goSorter.mbNum  = abNumber;
  goMnenhy.QuickSort(goSorter, 0, goTrainingDat.miShown - 1, abAscending);

  goMnenhy.UnsetWaitCursor(window);
}
