/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mnenhy.
 *
 * The Initial Developer of the Original Code is Karsten Düsterloh <mnenhy@tprac.de>.
 *
 * Portions created by the Initial Developer are Copyright (C) 2002-2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// +-
// |  the contents of this file will be loaded into the Mnenhy.installed.* context!
// |  ie. foo() is available as goMnenhy.installed.foo()
// +-

//
// todo: move the special chrome registry handling for goMnenhy.m_kbIsToolkitChromeRegistry into uninstall?
//


// |
// |  constants
// |

// debug
const gksInstalledDebugPref       = "extensions.mnenhy.log.installed";

// Mozilla applications (bit flags)
// - base bit flags
const kiMozillaBrandIDMozilla     = 1 << 0;   // includes SeaMonkey
const kiMozillaBrandIDThunderbird = 1 << 1;
const kiMozillaBrandIDFirefox     = 1 << 2;
// - composite flags
const kiMozillaBrandIDMail        = kiMozillaBrandIDMozilla | kiMozillaBrandIDThunderbird;
const kiMozillaBrandIDAll         = kiMozillaBrandIDMail    | kiMozillaBrandIDFirefox;

// Information about all existing Mnenhy subpackages:
const kiIndexAllowedApp  = 0; // bit flags: which applications use this package?
const kiIndexHasInitFile = 1; // bool: has an init file
const kiIndexInitContext = 2; // string: context to load init file into
const kiIndexInitFile    = 3; // string: init file URI
const koMnenhyPackages  =
    {"mnenhy-headers":    [kiMozillaBrandIDMail,    true, "headers", "chrome://mnenhy-headers/content/mnenhy-headers.js"],
     "mnenhy-codecs":     [kiMozillaBrandIDAll,     true, "codecs",  "chrome://mnenhy-codecs/content/mnenhy-codecs.js"  ],
     "mnenhy-chroman":    [kiMozillaBrandIDAll,     false],
     "mnenhy-regrig":     [kiMozillaBrandIDAll,     false],
     "mnenhy-junk":       [kiMozillaBrandIDMail,    false],
     "mnenhy-folderstore":[kiMozillaBrandIDMail,    true, "folderstore", "chrome://mnenhy-folderstore/content/mnenhy-folderstore.js"],
     "mnenhy-sidebar":    [kiMozillaBrandIDMozilla, false],
     "mnenhy":            [kiMozillaBrandIDAll,     false] };
const ksInstalledPrefix = "chrome://";
const ksInstalledSuffix = "/content/contents.rdf";
// Indices for GetAllPackages
const kiIndexInfoName     = 0; // string: package name
const kiIndexInfoAllowed  = 1; // bool: package allowed?
const kiIndexInfoInstall  = 2; // bool: package installed?
const kiIndexInfoActive   = 3; // bool: package active?


// |
// |  global variables
// |
var giMozillaApp     = kiMozillaBrandIDMozilla; // which Mozilla application is running?
var gbInstalledDebug = null;  // dump debug info

// |
// | functions
// |

// return an array of arrays ["package name" (, bIsAllowed, bIsInstalled, bIsActive)]
// data in () is only returned if abShowActive is true
function GetAllPackages(abShowActive)
{
  var rRet = [], rh, oh;
  for (var s in koMnenhyPackages)
  {
    rh = [];
    rh.push(s); // package name
    if (abShowActive)
    {
      oh = GetPackageStatus(s);
      rh.push(oh.bIsAllowed);   // package allowed?
      rh.push(oh.bIsInstalled); // package installed?
      rh.push(oh.bIsActive);    // package active?
    }
    rRet.push(rh);
  }
  return rRet;
}


// Is this Mnenhy package installed, i.e. registered with chrome? Is it active?
//  IN:     string  asPackage Mnenhy package name
//  RETURN: object  {bool bIsAllowed, bool bIsInstalled, bool bIsActive}
function GetPackageStatus(asPackage)
{
  var oRet = {bIsAllowed: false, bIsInstalled: false, bIsActive: false};
  if (asPackage in koMnenhyPackages)
  {
    // evil hack whilst the CM is not ready for the new chrome registry of FF/TB 1.5 (Gecko rv: 1.8)
    if ((asPackage == "mnenhy-chroman") && !IsMozilla() && (goMnenhy.ksMozillaVersion >= "1.8"))
    {
      goMnenhy.Log("installed.GetPackageStatus: no CM in Aviary 1.5: "+asPackage
                   +" -> Mozi="+IsMozilla()+" Version="+goMnenhy.ksMozillaVersion)
      oRet.bIsAllowed = false;
    }
    else
      oRet.bIsAllowed = !!(koMnenhyPackages[asPackage][kiIndexAllowedApp] & giMozillaApp);
      
    if (goMnenhy.m_kbIsToolkitChromeRegistry)
    { // chrome.rdf is no more, so interprete Mnenhy's chrome.manifest:
      // - installed packages appear there as a "content" line
      // - inactive packages appear there as comment-prefixed (#) "content" line
      const koMnenhyManifestFile = goMnenhy.uninstall.GetMnenhyManifestFile();
      if (koMnenhyManifestFile)
      {
        const ksMnenhyManifest = "\n" + goMnenhy.ReadFile(koMnenhyManifestFile) + "\n";
        // does the manifest file contain our package?
        const krsPackageKnown = ksMnenhyManifest.match(new RegExp("\\n\\s*(#?)\\s*content\\s+" + asPackage + "\\s"));
        debugdump("GetPackageStatus: krsPackageKnown="+krsPackageKnown);
        oRet.bIsInstalled = !!krsPackageKnown;  // have package
        oRet.bIsActive    = !!krsPackageKnown && (krsPackageKnown.length > 1) && (krsPackageKnown[1] == ""); // not a comment
      }
    }
    else
    { // read the activation state from the chrome.rdf
      oRet.bIsInstalled = goMnenhy.ExistURI(ksInstalledPrefix + asPackage + ksInstalledSuffix);
      if (oRet.bIsInstalled && goMnenhy.uninstall && goMnenhy.uninstall.mroChromeRDF)
      { // check if the 'disabled' arc exists in the chrome registry
        const koUnInst = goMnenhy.uninstall;
        const koRDF    = koUnInst.mroChromeRDF;
        // create the package URN
        const ksURN = koRDF[0].CreateURN([koUnInst.gksURNBasePack, asPackage], true);
        // look it up
        var oh;
        for (var i = 0; i < koRDF.length; ++i)
        {
          oh = koRDF[i];
          if (oh)
          {
            oRet.bIsActive = oh.GetItemValue("disabled", ksURN) != "true";
            if (!oRet.bIsActive)
              break;
          }
        }
      }
    }
  }
  debugdump("installed.GetPackageStatus for "+asPackage+" = "+oRet.toSource());
  return oRet;
}


function IsMozilla()
{
  return giMozillaApp == kiMozillaBrandIDMozilla;
}
function IsThunderbird()
{
  return giMozillaApp == kiMozillaBrandIDThunderbird;
}
function IsFirefox()
{
  return giMozillaApp == kiMozillaBrandIDFirefox;
}


// debug
function debugdump(aoDump)
{
  if (gbInstalledDebug)
    goMnenhy.Log(aoDump);
}


// |
// |  Init called by the global Mnenhy object when loading this script file
// |
function Init(aoMnenhy)
{ // keep pointer to have the Mnenhy object prototype available
  this.goMnenhy    = aoMnenhy;
  gbInstalledDebug = goMnenhy.GetPref(gksInstalledDebugPref, goMnenhy.kiPrefBool);
  
  // get the currently running application type
  if (    (goMnenhy.ksMozillaBrandName.search(/thunderbird/i) >= 0)
       || (goMnenhy.ksMozillaBrandName.search(/minotaur/i) >= 0))
    giMozillaApp = kiMozillaBrandIDThunderbird;
  else if (    (goMnenhy.ksMozillaBrandName.search(/firefox/i) >= 0)
            || (goMnenhy.ksMozillaBrandName.search(/firebird/i) >= 0))
    giMozillaApp = kiMozillaBrandIDFirefox;
  debugdump("installed.Init: giMozillaApp="+giMozillaApp);

  // load and init installed packages (if necessary)
  var rh, sh, oh, rsIllegalPackages = [];
  for (var s in koMnenhyPackages)
  {
    rh = koMnenhyPackages[s];
    // only check installed packages
    oh = GetPackageStatus(s);
    if (oh.bIsInstalled)
    {
      if (oh.bIsAllowed)
      { // init only allowed, active packages
        if (oh.bIsActive && rh[kiIndexHasInitFile])
        { // need to load and init
          sh = rh[kiIndexInitContext];  // get pseudo namespace name
          goMnenhy[sh] = {};            // create pseudo namespace
          if (goMnenhy.Include(rh[kiIndexInitFile], goMnenhy[sh], true))
            goMnenhy[sh].Init(aoMnenhy);
        } // need init
      }
      else
      { // Not allowed! If active, deactivate it.
        if (oh.bIsActive)
        {
          goMnenhy.Log("Package '"+s+"' not allowed for "+goMnenhy.ksMozillaBrandName+"!")
          goMnenhy.uninstall.SetAllowOverlaysForPackage(s, false);
          rsIllegalPackages.push(s);
        }
      }
    } // is installed
  } // for package init
  debugdump("installed.Init:: illegal packages="+rsIllegalPackages)
  
  // If we found illegal active packages, we need to restart.
  // (On second thought: do we really /need/ that?
  //  TB 0.6+ and FF 0.7+ do crash when trying and work perfectly without, so we disable it for them.
  //  TB/FF 1.5 with their new chrome registry need to restart, though.
  //  SeaMonkey doesnt need to disable packages anyway.)
  if (rsIllegalPackages.length && ((giMozillaApp & kiMozillaBrandIDMozilla) || goMnenhy.m_kbIsToolkitChromeRegistry) )
  {
    const ksIllegalPackages = goMnenhy.GetString("MnMsg_IllegalPackages")
                                      .replace("%p", rsIllegalPackages.join("', '"));
    goMnenhy.AlertService(ksIllegalPackages);
    // we need to invalidate the XUL.mfl to make the package deactivation happen!
    goMnenhy.uninstall.DeleteXulFastload();    
    if (!goMnenhy.GetPref("extensions.mnenhy.noshutdown", goMnenhy.kiPrefBool))
      goMnenhy.uninstall.goQuitApplication();
  }
}