/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mnenhy.
 *
 * The Initial Developer of the Original Code is Karsten Düsterloh <mnenhy@tprac.de>.
 *
 * Portions created by the Initial Developer are Copyright (C) 2002-2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// +-
// |  the contents of this file will be loaded into the Mnenhy.uninstall.* context!
// |  ie. foo() is available as goMnenhy.uninstall.foo()
// +-

// |
// |  constants
// |

// debug
const gksUninstallDebugPref      = "extensions.mnenhy.log.uninstall";
const gksUninstallNoBackupPref   = "extensions.mnenhy.uninstall.nobackup";
const gksUninstallNoShutdownPref = "extensions.mnenhy.noshutdown";


// |  CIDs
const gksCID_Fastload            = "@mozilla.org/fast-load-service;1";


// |  installed-chrome.txt entries
// package types
const gksTypeContent             = "content";
const gksTypeSkin                = "skin";
const gksTypeLocale              = "locale";
// package location
const gksLocInstall              = "install";
const gksLocProfile              = "profile";
const gksLocSelect               = "select";   // skin and locale only; usually once per i-c.txt
// package path uri type
const gksURIURL                  = "url";      // [jar:]resource: or [jar:]file:
const gksURIPath                 = "path";     // local path, eg. F:\\foo\\bar
// line ends
const gksCR                      = "\r";
const gksLF                      = "\n";
const gksCRLF                    = gksCR + gksLF;


// |  chrome.rdf URI parts
const gksURNBase                 = "urn"
const gksURNBaseCont             = "mozilla";
const gksURNBasePack             = "package";
const gksURNBaseSkin             = "skin";
const gksURNBaseLoc              = "locale";
const gksURNBaseRoot             = "root";
const gksURNBasePacks            = "packages";

// |  chrome.rdf node items
const gksItemPack                = "package";
const gksItemBaseURL             = "baseURL";
const gksItemLocType             = "locType";
//const gksItemSelSkin             = "selectedSkin";
//const gksItemSelLoc              = "selectedLocale";

// |  rdf meta data and file paths
const gksMozillaNamespace        = "http://www.mozilla.org/rdf/chrome#";
const gkrsMozillaBaseCont        = [gksURNBase, gksURNBaseCont];
const gksMozillaXUL_Fastload     = "XUL";
const gksMozillaChrome           = "chrome/";
const gksMozillaExtensions       = "extensions/";
const gksMozillaChromeRDF        = "chrome.rdf";
const gksMozillaInstChromeTxt    = "installed-chrome.txt";
const gksMozillaOverlayInfoDir   = "overlayinfo";
const gksMozillaOverlayInfo      = gksMozillaOverlayInfoDir + "/";
const gksMozillaOverlaysRDF      = "overlays.rdf";
const gksMozillaOverlaysRDFPath  = "/content/" + gksMozillaOverlaysRDF;
const gksMozillaStylesheetsRDF   = "stylesheets.rdf";
const gksMozillaStylesheetsRDFPath = "/skin/" + gksMozillaStylesheetsRDF;
const gksMozillaOverlaysRoot     = "urn:mozilla:overlays"
const gksMozillaStylesheetsRoot  = "urn:mozilla:stylesheets"
const gksMozillaContentsRDF      = "contents.rdf";
const gksBackupExtension         = ".mnenhy.bak";

// copied from CM
const gksFlagJarScheme           = "jar:";
const gksFlagJarMidfix           = ".jar!";


// |  god mode test
const gksGodMode                 = "GodMode.txt";


// |  global prefs (see ReadPrefs)
var gbUninstallDebug             = null;  // dump debug info
var gbNoBackup                   = null;  // don't make backup
var gbNoShutdown                 = null;  // don't shut down


// |
// |  main functions
// |  
// |  Uninstall:                convenience function; will call several of the functions below
// |  UninstallPackages:        convenience function; will call several of the functions below
// |
// |  UninstallContentPackage:  uninstall the given content package from both user chrome
// |                            and app chrome (if possible) and remove all its files stored as
// |                            as baseURLs (if possible)
// |  UninstallSkinPackage:     as above, but for skin packages
// |  UninstallLocalePackage:   as above, but for locale packages
// |
// |
// |  BackupChromeRegistry:     backup all Mozilla system files possibly changed during uninstall
// |                            by renaming them to <name>.mnenhy.bak
// |                            user: chrome.rdf, overlayinfo
// |                            app:  chrome.rdf, overlayinfo, installed-chrome.txt
// |                            THIS FUNCTION SHOULD BE CALLED PRIOR TO ANY UNINSTALLING!
// |  RestoreChromeRegistry:    restore the files backupped via BackupChromeRegistry
// |
// |
// |  RemoveDeletableFiles:     delete package files that aren't used anymore
// |                            (this CANNOT be undone!)
// |  PurgeChromeRegistry:      delete XUL.mfasl, user overlayinfo and app overlayinfo,
// |                            rebuild both overlayinfo
// |  ShutdownMozilla:          shutdown Mozilla



// Uninstall a package of the given type, but backup data before and clean up afterwards
// This function will shutdown Mozilla!
//  IN: array arPacks        -> an array of arrays of strings
//                              arName contains an entry for every package,
//                              every package is an array [package name, package type]
//                              (package type: see gksType*, defaults to gksTypeContent)
//                              eg. Uninstall([["foo", gksTypeSkin],["bar"]])
//                              will uninstall the skin package "foo" and the content package "bar"
function Uninstall(arPacks)
{
  debugdump("Uninstall+");
  if (arPacks)
  { // make backup
    BackupChromeRegistry();
    // collect deletable files
    var oDeletableFiles = {};
    // uninstall by type
    UninstallPackages(arPacks, oDeletableFiles);
    // clean-up
    PurgeChromeRegistry();
    // shutdown and file deletion
    ShutdownMozilla(oDeletableFiles, true);
  }
  debugdump("Uninstall-");
}


// Uninstall a package of the given type
//  IN:     array arPacks           -> an array of arrays of strings
//                                     arName contains an entry for every package,
//                                     every package is an array [package name, package type]
//                                     (package type: see gksType*, defaults to gksTypeContent)
//                                     eg. Uninstall([["foo", gksTypeSkin],["bar"]]) will uninstall
//                                     the skin package "foo" and the content package "bar"
//  IN/OUT: object aoDeletableFiles -> collects the URIs that can be deleted by the caller
function UninstallPackages(arPacks, aoDeletableFiles)
{
  debugdump("UninstallPackages+");
  if (arPacks)
  { // collect deletable files
    if (!aoDeletableFiles)
      aoDeletableFiles = {};
    // uninstall by type
    for (var iPack in arPacks)
    {
      const koPack = arPacks[iPack];
      const ksName = (koPack && (0 in koPack)) ? koPack[0] : "";
      const ksType = (koPack && (1 in koPack)) ? koPack[1] : "";

      switch(ksType)
      {
        case gksTypeSkin:   UninstallSkinPackage   (ksName, aoDeletableFiles); break;
        case gksTypeLocale: UninstallLocalePackage (ksName, aoDeletableFiles); break;
        case gksTypeContent:
        default:            UninstallContentPackage(ksName, aoDeletableFiles); break;
      }
    }
  }
  debugdump("UninstallPackages-");
}


// return a new file object that represents the backup location for aoFile
function CreateBackupLocation(aoFile)
{
  const koPH    = goMnenhy.GetProtocolHandler();
  const ksURI   = koPH.getURLSpecFromFile(aoFile);
  const koParts = ChromeURI2Info(ksURI);
  const ksBackupURI = ChromeInfo2URI(ksURI,
                                     koParts.bProfile,
                                     koParts.bJar,
                                     koParts.sType,
                                     koParts.bExtension,
                                     koParts.sUUID,
                                     gksBackupExtension + "/" + koParts.sPackage,
                                     koParts.sInternal);
  const koBak = koPH.getFileFromURLSpec(ksBackupURI);
  return koBak;
}


// Backup all Mozilla system files possibly changed during uninstall
// by moving them to .mnenhy.bak/*
//  user: chrome.rdf, overlayinfo
//  app:  chrome.rdf, overlayinfo, installed-chrome.txt
//
// XXX TODO MNYR: save extensions/Extensions.rdf?
//
function BackupChromeRegistry()
{
  debugdump("BackupChromeRegistry+");
  ReadPrefs();
  if (gbNoBackup)
    debugdump("BackupChromeRegistry disabled!");
  else
  { // user files
    var oChromeDirUserBak = goMnenhy.m_koChromeDirUser.clone();
    oChromeDirUserBak.append(gksBackupExtension);
    goMnenhy.CopyFile(goMnenhy.m_koChromeDirUser,    gksMozillaChromeRDF,
                                  oChromeDirUserBak, gksMozillaChromeRDF,      true);
    goMnenhy.CopyFile(goMnenhy.m_koChromeDirUser,    gksMozillaOverlayInfoDir,
                                  oChromeDirUserBak, gksMozillaOverlayInfoDir, true);
    // bug 278534 "flattens" the overlayinfo into single files (2005-02-23)
    goMnenhy.CopyFile(goMnenhy.m_koChromeDirUser,    gksMozillaOverlaysRDF,
                                  oChromeDirUserBak, gksMozillaOverlaysRDF,    true);
    goMnenhy.CopyFile(goMnenhy.m_koChromeDirUser,    gksMozillaStylesheetsRDF,
                                  oChromeDirUserBak, gksMozillaStylesheetsRDF, true);
    // app files
    if (this.mbIsGodMode)
    { // only if we're allowed to do so
      debugdump("BackupChromeRegistry++");
      var oChromeDirAppBak = goMnenhy.m_koChromeDirApp.clone();
      oChromeDirAppBak.append(gksBackupExtension);
      goMnenhy.CopyFile(goMnenhy.m_koChromeDirApp,    gksMozillaChromeRDF,
                                    oChromeDirAppBak, gksMozillaChromeRDF,      true);
      goMnenhy.CopyFile(goMnenhy.m_koChromeDirApp,    gksMozillaOverlayInfoDir,
                                    oChromeDirAppBak, gksMozillaOverlayInfoDir, true);
      goMnenhy.CopyFile(goMnenhy.m_koChromeDirApp,    gksMozillaInstChromeTxt,
                                    oChromeDirAppBak, gksMozillaInstChromeTxt,  true);
      // bug 278534 "flattens" the overlayinfo into single files (2005-02-23)
      goMnenhy.CopyFile(goMnenhy.m_koChromeDirApp,    gksMozillaOverlaysRDF,
                                    oChromeDirAppBak, gksMozillaOverlaysRDF,    true);
      goMnenhy.CopyFile(goMnenhy.m_koChromeDirApp,    gksMozillaStylesheetsRDF,
                                    oChromeDirAppBak, gksMozillaStylesheetsRDF, true);
      debugdump("BackupChromeRegistry--");
    }
  }
  debugdump("BackupChromeRegistry-");
}

// Restore the files backupped via BackupChromeRegistry
//  user: chrome.rdf, overlayinfo
//  app:  chrome.rdf, overlayinfo, installed-chrome.txt
//
// XXX TODO MNYR: restore extensions/Extensions.rdf?
//
function RestoreChromeRegistry()
{
  debugdump("RestoreChromeRegistry+");
  // user files
  var oChromeDirUserBak = goMnenhy.m_koChromeDirUser.clone();
  oChromeDirUserBak.append(gksBackupExtension);
  goMnenhy.CopyFile(            oChromeDirUserBak, gksMozillaChromeRDF,
                    goMnenhy.m_koChromeDirUser,    gksMozillaChromeRDF,      true);
  goMnenhy.CopyFile(            oChromeDirUserBak, gksMozillaOverlayInfoDir,
                    goMnenhy.m_koChromeDirUser,    gksMozillaOverlayInfoDir, true);
  // bug 278534 "flattens" the overlayinfo into single files (2005-02-23)
  goMnenhy.CopyFile(            oChromeDirUserBak, gksMozillaOverlaysRDF,
                    goMnenhy.m_koChromeDirUser,    gksMozillaOverlaysRDF,    true);
  goMnenhy.CopyFile(            oChromeDirUserBak, gksMozillaStylesheetsRDF,
                    goMnenhy.m_koChromeDirUser,    gksMozillaStylesheetsRDF, true);
  // app files
  if (this.mbIsGodMode)
  { // only if we're allowed to do so
    debugdump("RestoreChromeRegistry++");
    var oChromeDirAppBak = goMnenhy.m_koChromeDirApp.clone();
    oChromeDirAppBak.append(gksBackupExtension);
    goMnenhy.CopyFile(            oChromeDirAppBak, gksMozillaChromeRDF,
                      goMnenhy.m_koChromeDirApp,    gksMozillaChromeRDF,      true);
    goMnenhy.CopyFile(            oChromeDirAppBak, gksMozillaOverlayInfoDir,
                      goMnenhy.m_koChromeDirApp,    gksMozillaOverlayInfoDir, true);
    goMnenhy.CopyFile(            oChromeDirAppBak, gksMozillaInstChromeTxt,
                      goMnenhy.m_koChromeDirApp,    gksMozillaInstChromeTxt,  true);
    // bug 278534 "flattens" the overlayinfo into single files (2005-02-23)
    goMnenhy.CopyFile(            oChromeDirAppBak, gksMozillaOverlaysRDF,
                      goMnenhy.m_koChromeDirApp,    gksMozillaOverlaysRDF,    true);
    goMnenhy.CopyFile(            oChromeDirAppBak, gksMozillaStylesheetsRDF,
                      goMnenhy.m_koChromeDirApp,    gksMozillaStylesheetsRDF, true);
    debugdump("RestoreChromeRegistry--");
  }
  debugdump("RestoreChromeRegistry-");
}


// The members of the aoFiles object are URIs of the files to be deleted
function RemoveDeletableFiles(aoFiles, abBackup)
{
  debugdump("RemoveDeletableFiles+");
  if (aoFiles)
  { // we need to release all cached jar files before deleting them
    // because Win2k (and maybe others) don't allow for deletion of open files
    const koJarProtHandler = goMnenhy.GetProtocolHandler(goMnenhy.ksProtocolJar);
    var oCache = null;
    if (koJarProtHandler instanceof Components.interfaces.nsIJARProtocolHandler)
      oCache = koJarProtHandler.JARCache;
    for (var sFile in aoFiles)
    {
      const koFile = ConvertURI2File(sFile);
      if (koFile.exists())
      { // release it
        try
        {
          if (oCache && koFile.isFile())
          {
            const koZipReader = oCache.getZip(koFile);
            if (koZipReader)
              koZipReader.close();
          }
          // if requested, make backup copies in the local .mnenhy.bak directory
          if (abBackup)
          {
            const koBak = CreateBackupLocation(koFile);
            goMnenhy.Log("RemoveDeletableFiles: Copying " + koFile.path + " to " + koBak.path);
            goMnenhy.CopyFile(koFile.parent, koFile.leafName, koBak.parent, koBak.leafName);
          }
          goMnenhy.Log("RemoveDeletableFiles: Deleting " + koFile.path);
          koFile.remove(true);
        }
        catch(oError)
        {
          goMnenhy.Log("Error in mnenhy-uninstall::remove deletable files(" + koFile.path + "):\n"
               + oError);
          return oError;
        }
      } // found file
    } // loop over files
  } // have files
  debugdump("RemoveDeletableFiles-");
}


/*
// The members of the aoFiles object are URIs of the files to be deleted
function RemoveDeletableFiles(aoFiles)
{ // Empty the nsIZipReaderCache of nsJar by faking a low memory condition
  // If we wouldn't do that, Win2k would block the deletion of files still in the cache
  debugdump("RemoveDeletableFiles+");
  if (aoFiles)
  {
    const gksCID_ObserverService = "@mozilla.org/observer-service;1";
    var oObserverService = Components.classes[gksCID_ObserverService]
                                     .getService(Components.interfaces.nsIObserverService);
    if (oObserverService)
    { // send fake memory problem
      oObserverService.notifyObservers(null, "memory-pressure", "Mnenhy!");
      // delete the files
      for (var sFile in aoFiles)
      {
        const koFile = ConvertURI2File(sFile);
        if (koFile.exists())
        { // release it
          try
          {
            debugdump("Deleting " + koFile.path);
            koFile.remove(true);
          }
          catch(oError)
          {
            goMnenhy.Log("Error in mnenhy-uninstall::remove deletable files(" + koFile.path + "):\n"
                 + oError);
            return oError;
          }
        } // found file
      } // loop over files
    } // have service
  } // have files
  debugdump("RemoveDeletableFiles-");
}
*/


// Delete XUL.mfasl, user overlayinfo and app overlayinfo, rebuild both overlayinfo
function PurgeChromeRegistry()
{ // user files
  debugdump("PurgeChromeRegistry+");
  DeleteXulFastload();
  goMnenhy.DeleteFile(goMnenhy.m_koChromeDirUser, gksMozillaOverlayInfoDir);
  // bug 278534 "flattens" the overlayinfo into single files (2005-02-23)
  goMnenhy.DeleteFile(goMnenhy.m_koChromeDirUser, gksMozillaOverlaysRDF);
  goMnenhy.DeleteFile(goMnenhy.m_koChromeDirUser, gksMozillaStylesheetsRDF);
  UpdateAllOverlayInfo(true);
  // app files
  if (this.mbIsGodMode)
  { // only if we're allowed to do so
    debugdump("PurgeChromeRegistry++");
    goMnenhy.DeleteFile(goMnenhy.m_koChromeDirApp, gksMozillaOverlayInfoDir);
    // bug 278534 "flattens" the overlayinfo into single files (2005-02-23)
    goMnenhy.DeleteFile(goMnenhy.m_koChromeDirApp, gksMozillaOverlaysRDF);
    goMnenhy.DeleteFile(goMnenhy.m_koChromeDirApp, gksMozillaStylesheetsRDF);
    UpdateAllOverlayInfo(false);
    debugdump("PurgeChromeRegistry--");
  }
  debugdump("PurgeChromeRegistry-");
}


// delete all deletable files, close all windows, shutdown Mozilla
function ShutdownMozilla(aoDeletableFiles, abBackup, asPrefsBranch)
{
  ReadPrefs();
  debugdump("ShutdownMozilla+");

  // remove the remainder of the uninstalled packages
  CreateDeletableBaseURLs(mroChromeRDF[0], aoDeletableFiles);
  if (mroChromeRDF[1])
    CreateDeletableBaseURLs(mroChromeRDF[1], aoDeletableFiles);

// Alternative method to call RemoveDeletableFiles
// (Not fully tested and not really better - the problem is goQuitApplication()!
//  Closing Mozilla via the window's upper right [x] button will do nicely,
//  but goQuitApplication() makes it choke... :-| )
//
//const gksCID_ObserverService = "@mozilla.org/observer-service;1";
//var oObserverService = Components.classes[gksCID_ObserverService]
//                                 .getService(Components.interfaces.nsIObserverService);
//if (oObserverService)
//{
//  var obstest = {observe: function(aoSubjectIgnored, asTopicIgnored, asStateIgnored)
//                          {
//                            this.RemoveDeletableFiles(this.aoDeletableFiles, this.abBackup);
//                          },
//                 RemoveDeletableFiles: RemoveDeletableFiles,
//                 oDeletableFiles: aoDeletableFiles,
//                 bBackup: abBackup
//                 };
//  oObserverService.addObserver(obstest, "quit-application", false);
//}

  RemoveDeletableFiles(aoDeletableFiles, abBackup);

  // the following call to goQuitApplication will dump
  // an obviously(?) bogus(?) error to the console
  // [should be fixed by using goQuitApplicationTimed?]
  if (!gbNoShutdown)
    this.goQuitApplication();
  debugdump("ShutdownMozilla-");
}


// Uninstall a content package:
//  IN:     string asName           -> the name of the content package, eg. 'mnenhy'
//  IN/OUT: object aoDeletableFiles -> all deletable files are added as object members
//
// Uninstall a content package by
// - removing all skins/locales "urn:mozilla:<skin|locale>:<skin|lang>:<asName>"
//   incl. their installed-chrome.txt lines if applicable
//   (and adding their baseURLs to aoBaseURLs)
// - removing the content "urn:mozilla:package:<asName>"
//   incl. its installed-chrome.txt line if applicable
//   (and adding its baseURLs to aoBaseURLs)
// The files in aoBaseURLs can be deleted by the caller.
function UninstallContentPackage(asName, aoDeletableFiles)
{
  debugdump("UninstallContentPackage("+asName+")+");
  if (asName && aoDeletableFiles)
  { // content pack URNs
    const krContent = [gksURNBasePack, asName];
    // loop over chrome.rdf files
    for (var iChrome = 0; (iChrome < mroChromeRDF.length) && (mroChromeRDF[iChrome]); ++iChrome)
    {
      var oChromeRDF   = mroChromeRDF[iChrome];
      var oElementData = {oChromeRDF: oChromeRDF,
                          sICText:    "",
                          bIsProfile: !iChrome,
                          oBaseURLs:  {} };
      var rNode;
      // open installed-chrome.txt (if in GodMode)
      if (this.mbIsGodMode)
        oElementData.sICText = LoadInstalledChrome();
      // process skin/locales
      const koSkinsAndLocales = oChromeRDF.GetSourcesForNode(gksItemPack, krContent);
      while (koSkinsAndLocales.hasMoreElements())
      { // read the baseURL from the skin or locale package entry before deleting it
        var oSkinOrLocale = koSkinsAndLocales.getNext()
        // get baseURL, delete installed-chrome.txt entry
        UninstallPackage(oSkinOrLocale, oElementData);
        // remove node, eg urn:mozilla:locale:de-DE:mnenhy
        rNode = oSkinOrLocale.Value.split(":").slice(2);
        oChromeRDF.RemoveNode(rNode);
        // if parent container is emtpy now, remove it, too
        rNode.pop();
        rNode.push(gksURNBasePacks);
        if (!oChromeRDF.GetContainerSize(rNode))
        { // remove container node, eg urn:mozilla:locale:de-DE:packages
          oChromeRDF.RemoveContainer(rNode);
          // and the containing <description>, eg. urn:mozilla:locale:de-DE
          rNode.pop();
          oChromeRDF.RemoveNode(rNode);
          // if this parent container is emtpy now also, remove it
          // (that does mean removing the skin/locale root!)
          rNode = [gksURNBasePack, gksURNBaseRoot];
          if (!oChromeRDF.GetContainerSize(rNode))
          { // remove the root node
            oChromeRDF.RemoveContainer(rNode);
          }
        }
      } // loop over skins and locales
      // process main element
      const koContent = oChromeRDF.ExistNode(krContent);
      if (koContent)
      { // remember content node baseURL
        UninstallPackage(koContent, oElementData);
        // remove node
        rNode = koContent.Value.split(":").slice(2);
        oChromeRDF.RemoveNode(rNode);
      }
      // save changes to installed-chrome.txt
      if (this.mbIsGodMode)
        SaveInstalledChrome(oElementData.sICText);
      // collect oBaseURLs
      for (var sBaseURL in oElementData.oBaseURLs)
        aoDeletableFiles[sBaseURL] = oElementData.oBaseURLs[sBaseURL];
    } // for loop over chrome.rdf files
  } // have args
  debugdump("UninstallContentPackage-");
}


// Uninstall a skin or locale package:
//  IN:     string asProvider       -> the package type: gksURNBaseSkin or gksURNBaseLoc
//  IN:     string asName           -> the name of the package, eg. 'modern/1.0' or 'de-AT'
//  IN/OUT: object aoDeletableFiles -> deletable files will be added as object members
//
// Uninstall a skin or locale package by
// - deselecting it from all content packages
// - removing all skins/locales "urn:mozilla:<skin|locale>:<asName>:*"
//   incl. their installed-chrome.txt lines if applicable
//   (and adding their baseURLs to aoBaseURLs)
// - removing the skin/locale "urn:mozilla:<skin|locale>:<asName>"
//   incl. its installed-chrome.txt line if applicable
// The files in aoDeletableFiles can be deleted by the caller.
function UninstallSkinPackage(asName, aoDeletableFiles)
{
  debugdump("UninstallSkinPackage+");
  UninstallSkinOrLocalePackage(gksURNBaseSkin, asName, aoDeletableFiles);
  debugdump("UninstallSkinPackage-");
}
function UninstallLocalePackage(asName, aoDeletableFiles)
{
  debugdump("UninstallLocalePackage+");
  UninstallSkinOrLocalePackage(gksURNBaseLoc, asName, aoDeletableFiles);
  debugdump("UninstallLocalePackage-");
}
function UninstallSkinOrLocalePackage(asProvider, asName, aoDeletableFiles)
{
  debugdump("UninstallSkinOrLocalePackage("+asProvider+","+asName+")+");
  var fDeselector = null;
  switch (asProvider)
  {
    case gksURNBaseSkin:  fDeselector = goMnenhy.m_koChromeRegService.deselectSkin;   break;
    case gksURNBaseLoc:   fDeselector = goMnenhy.m_koChromeRegService.deselectLocale; break;
  }

  if (fDeselector && asName && aoBaseURLs)
  { // skin/locale pack URNs
    const krSkinOrLoc = [asProvider, asName];
    // loop over chrome.rdf files
    for (var iChrome = 0; (iChrome < mroChromeRDF.length) && (mroChromeRDF[iChrome]); ++iChrome)
    {
      var oChromeRDF   = mroChromeRDF[iChrome];
      var oElementData = {oChromeRDF: oChromeRDF,
                          sICText:    "",
                          bIsProfile: !iChrome,
                          oBaseURLs:  {} };
      // open installed-chrome.txt (if in GodMode)
      if (this.mbIsGodMode)
        oElementData.sICText = LoadInstalledChrome();
      // try to deselect skin/locale
      try
      {
        fDeselector(asName, !iChrome);
      }
      catch(ignored){}
      // process skin/locales
      var oSkinOrLocPacks = krSkinOrLoc.slice();
      oSkinOrLocPacks.push(gksURNBasePacks);
      oChromeRDF.RemoveContainer(oSkinOrLocPacks, UninstallPackage, oElementData);
      // remove the skin/locales main <rdf:description>
      oChromeRDF.RemoveNode(krSkinOrLoc);
      // save changes to installed-chrome.txt
      if (this.mbIsGodMode)
        SaveInstalledChrome(oElementData.sICText);
      // collect oBaseURLs
      for (var sBaseURL in oElementData.oBaseURLs)
        aoDeletableFiles[sBaseURL] = oElementData.oBaseURLs[sBaseURL];
    } // loop over chrome.rdf files
  } // have valid arguments
  debugdump("UninstallSkinOrLocalePackage-");
}


// Get the baseURL property of the nsIRDFResource aoElement and
// uninstall the corresponding installed-chrome.txt line (if possible).
// aoElementData must provide these members:
//    oChromeRDF: the chrome.rdf MnenhyRDF object to work with
//    oBaseURLs:  the storage member for the baseURLs
//    bIsProfile: is oChromeRDF the user profile chrome.rdf?
//    sICText:    the text content of the installed-chrome.txt
function UninstallPackage(aoElement, aoElementData)
{ // read the baseURL from the skin or locale package entry before deleting it
  debugdump("UninstallPackage+");
  if (aoElement instanceof goMnenhy.rdf.gkIRDFResource)
  { // create node urn
    var rNode = aoElement.Value.split(":").slice(2);
    // read baseURL
    var sBaseURL = aoElementData.oChromeRDF.GetItemValue(gksItemBaseURL, rNode);
    if (sBaseURL)
    {
      aoElementData.oBaseURLs[sBaseURL] = aoElementData.bIsProfile;
      if (this.mbIsGodMode)
      { // delete installed-chrome.txt line (we don't know about path!)
        var sLine = CreateICLine(sBaseURL, rNode[0], aoElementData.bIsProfile, null, "\\s");
        var iSize = aoElementData.sICText.length;
        aoElementData.sICText = aoElementData.sICText.replace(new RegExp(sLine), "\n");
        if (aoElementData.sICText.length == iSize)
          goMnenhy.Log("UninstallPackage:Warning: couldn't find installed-chrome.txt line: "+sLine);
      } // app chrome
    }
  } // have nsIRDFResource
  debugdump("UninstallPackage-");
}



// |
// |  uninstall helper functions
// |

// Get chrome information from URI
//
// jar:resource:/chrome/mnenhy.jar!/content/mnenhy/...
//  |  \_______________/  |   \__/  \_____/\_________/
//  |          |          |     |    sType sInternal
//  |          |          |     |  \_________________/
//  |          |          |     |          |
//  |     sChromePath     |     |      sEntryPath
// sJarPrefix          sPackage `sJarSuffix
//
// For profile installed packages, sChromePath is usually a file: URI.
// For packages installed via the Extension Manager, sChromePath has this general structure:
//    scheme:/extensions/$UUID/chrome
//
// IN:      string  asURI   The URI to split
// RETURN:  {
//            bool   bJar,        // if |true|, the URI points into a .jar file
//            bool   bProfile,    // if |true|, the URI points into the user profile
//            bool   bExtension,  // if |true|, the URI points into /extensions
//            string sUUID,       // the package UUID (if any)
//            string sPackage,    // the package name
//            string sType,       // the package type
//            string sInternal    // rest of the entry path, usually with leading /
//          }
function ChromeURI2Info(asURI)
{
  debugdump("ChromeURI2Info("+asURI+")+");
  var oRet = {
               bJar:       null,
               bProfile:   null,
               bExtension: null,
               sUUID:      null,
               sPackage:   "",
               sType:      "",
               sInternal:  ""
             };
  if (asURI)
  { // create URI from string
    var oURI = goMnenhy.m_koIOService.newURI(asURI, null, null);
    // handle jar: URIs
    oRet.bJar = oURI instanceof Components.interfaces.nsIJARURI;
    if (oRet.bJar && oURI.JARFile instanceof Components.interfaces.nsIURL)
    { // drop the 'jarity'
      oURI = goMnenhy.m_koIOService.newURI(   oURI.JARFile.prePath + oURI.JARFile.directory
                                            + oURI.JARFile.fileBaseName + "/" + oURI.JAREntry,
                                            null, null);
    }
    // canonify URI to file: URI
    oURI = goMnenhy.ConvertURI2FileURI(oURI);
    // check profile
    oRet.bProfile = !oURI.spec.search(goMnenhy.m_ksURIUser);
    // clean location from asURI
    asURI = oURI.spec.replace(oRet.bProfile ? goMnenhy.m_ksURIUser : goMnenhy.m_ksURIApp, "");
    // check aviarity
    oRet.bExtension = !asURI.search("^" + gksMozillaExtensions);
    if (oRet.bExtension)
    {
      const krUUID = asURI.match(gksMozillaExtensions + "([^/]+)/");
      if (krUUID && krUUID.length > 1)
      {
        oRet.sUUID = krUUID[1];
        asURI = asURI.replace(krUUID[0], "");
      }
    }
    // asURI should now be in this format: chrome/$sPackage/$sType$sInternal
    // (the variables $sType and $sInternal may be empty, eg. for directory paths)
    const krParts = asURI.match(gksMozillaChrome + "([^/]+)(?:/([^/]+)(.+))?");
    if (krParts)
    {
      switch (krParts.length)
      { // fall through cases; at least create empty strings
        case 4: oRet.sInternal = krParts[3] || "";
        case 3: oRet.sType     = krParts[2] || "";
        case 2: oRet.sPackage  = krParts[1] || "";
      }
    }
  }
  debugdump("ChromeURI2Info-");
  return oRet;
}


// Build URI from chrome info
//
// IN:     string  asURI        URI to be changed according to the other parameters
// IN:     bool    abProfile    if |true|, create profile URI
// IN:     bool    abJar        if |true|, create jar URI
// IN:     string  asType       'content', 'skin' or 'locale'
// IN:     bool    abExtension  if |true|, use /extensions instead of /chrome
// IN:     string  asUUID:      new UUID if abExtension is |true|
// IN:     string  asPackage:   new package name
// IN:     string  asInternal:  new internal path
// RETURN: string             the new URI
//
// Parameters with value |null| will be ignored!
//
// XXX TODO : should we update all the other EM data?
// XXX TODO : Eg Extensions.rdf, installed-extensions-processed.txt and Uninstall?
// XXX TODO : Actually, yes, but not here
function ChromeInfo2URI(asURI, abProfile, abJar, asType, abExtension, asUUID, asPackage, asInternal)
{
  debugdump("ChromeInfo2URI("+asURI      +","
                             +abProfile  +","
                             +abJar      +","
                             +asType     +","
                             +abExtension+","
                             +asUUID     +","
                             +asPackage  +","
                             +asInternal +")+");
  var sNewURI = asURI;
  if (sNewURI)
  { // x != null is needed to allow x == 0 and x == ""
    var oInfo = ChromeURI2Info(sNewURI);
    if (abProfile != null)
      oInfo.bProfile = abProfile;
    if (abJar != null)
      oInfo.bJar = abJar;
    if (asType != null)
      oInfo.sType = asType;
    if (abExtension != null)
      oInfo.bExtension = abExtension;
    if (asUUID != null)
      oInfo.sUUID = asUUID;
    if (asPackage != null)
      oInfo.sPackage = asPackage;
    if (asInternal != null)
      oInfo.sInternal = asInternal;

    // create the new base package URI
    sNewURI = oInfo.bProfile ? goMnenhy.m_ksURIUser : goMnenhy.m_ksURIApp;
    if (oInfo.bExtension)
    {
      if (!oInfo.sUUID)
      { // fake the UUID
        // XXX TODO MNYR: should we try to read the UUID of oInfo.sPackage from RDF?
        oInfo.sUUID = oInfo.sPackage;
      }
      if (oInfo.sUUID)
      {
        sNewURI += gksMozillaExtensions + oInfo.sUUID + "/";
      }
    }
    sNewURI += gksMozillaChrome + oInfo.sPackage;
    // enjar it if requested
    if (oInfo.bJar)
      sNewURI = gksFlagJarScheme + sNewURI + gksFlagJarMidfix;
    // add type and internal path
    sNewURI += "/" + oInfo.sType + oInfo.sInternal;
  }
  debugdump("ChromeInfo2URI["+sNewURI+"]-");
  return sNewURI;
}


// Check which of the baseURLs in the associative array aoBaseURLs
// are still in use in chrome.rdf aoChromeRDF.
// All baseURLs still in use will be deleted from aoBaseURLs.
// The elements of aoBaseURLs will be converted into nsIURI spec strings,
// nsIJARURIs will be replaced by their JARFile URIs.
// After runnning this functions, all files left in aoBaseURLs can be deleted!
function CreateDeletableBaseURLs(aoChromeRDF, aoBaseURLs)
{
  debugdump("CreateDeletableBaseURLs+");
  if (aoChromeRDF && aoBaseURLs)
  { // prepare aoBaseURLs
    var iLength = 0;
    for (var sBaseURL in aoBaseURLs)
    {
      var oURI = goMnenhy.m_koIOService.newURI(sBaseURL, null, null);
      if (oURI instanceof Components.interfaces.nsIJARURI)
        oURI = oURI.JARFile;
      if (oURI.spec != sBaseURL)
      { // convert URL to URI
        aoBaseURLs[oURI.spec] = aoBaseURLs[sBaseURL];
        delete aoBaseURLs[sBaseURL];
      }
      ++iLength;
    }
    // check it!
    const koBaseURLRes = aoChromeRDF.GetItemRes(gksItemBaseURL);
    const koAllRes     = aoChromeRDF.moDataSource.GetAllResources();
    while (koAllRes.hasMoreElements() && iLength)
    { // inspect every resource
      const koRes = koAllRes.getNext();
      if (koRes instanceof goMnenhy.rdf.gkIRDFResource)
      { // check all baseURL arcs
        const koAllBaseURLs = aoChromeRDF.moDataSource.GetTargets(koRes, koBaseURLRes, true);
        while (koAllBaseURLs.hasMoreElements() && iLength)
        {
          const koBaseURL = koAllBaseURLs.getNext();
          if (koBaseURL instanceof goMnenhy.rdf.gkIRDFLiteral)
          { // using URIs makes JAR testing very easy
            var oURI = goMnenhy.m_koIOService.newURI(koBaseURL.Value, null, null);
            if (oURI instanceof Components.interfaces.nsIJARURI)
              oURI = oURI.JARFile;
            // check our baseURL array
            for (var sBaseURI in aoBaseURLs)
            {
              if (!oURI.spec.search(sBaseURI))
              { // still in use, mustn't delete that file
                delete aoBaseURLs[sBaseURI];
                --iLength;
              }
            } // for all aoBaseURLs
          } // baseURL is literal
        } // while all baseURLs
      } // resource is nsIRDFRresource
    } // while all resources
  } // valid parameters
  debugdump("CreateDeletableBaseURLs-");
}


// loop over all packages in chrome.rdf and rebuild their overlayinfo
function UpdateAllOverlayInfo(abUser)
{
  debugdump("UpdateAllOverlayInfo+");
  const koChromeRDF = mroChromeRDF[abUser ? 0 : 1]
  if (koChromeRDF)
  { // save and reload chrome.rdf
    koChromeRDF.FlushDataSource();

    // init the overlay cache
    // this will cache the open overlays.rdf and stylesheets.rdf
    // it's an associative array with path:object entries
    var oOverlayCache = {}

    // loop over content packages to generate overlayinfo
    var rsPack, sChromeURI, i;
    var rsPacks = koChromeRDF.GetNodeList([gksURNBasePack, gksURNBaseRoot]);
    if (rsPacks)
    {
      for (i in rsPacks)
      {
        rsPack = rsPacks[i].split(":").slice(2);
        if (rsPack)
        { // first read the package contents.rdf
          sChromeURI = koChromeRDF.GetItemValue(gksItemBaseURL, rsPack);
          if (sChromeURI)
            UpdateOverlayInfo(oOverlayCache, sChromeURI + gksMozillaContentsRDF, abUser);
        } // have urn
      } // for all content packages
    } // have content packages

    // loop over skin packages to generate overlayinfo
    const krsSkins = koChromeRDF.GetNodeList([gksURNBaseSkin, gksURNBaseRoot]);
    if (krsSkins)
    { // loop over all skin packages
      for (var j in krsSkins)
      { // get the content packages of this skin package
        const krsSkin = krsSkins[j].split(":").slice(2);
        const ksPacks = koChromeRDF.GetItemValue(gksURNBasePacks, krsSkin);
        if (ksPacks)
        {
          const krsPacks = ksPacks.split(":").slice(2);
          rsPacks = koChromeRDF.GetNodeList(krsPacks);
          // loop over all packages of this skin
          for (i in rsPacks)
          {
            rsPack = rsPacks[i].split(":").slice(2);
            if (rsPack)
            { // first read the package contents.rdf
              sChromeURI = koChromeRDF.GetItemValue(gksItemBaseURL, rsPack);
              if (sChromeURI)
                UpdateOverlayInfo(oOverlayCache, sChromeURI + gksMozillaContentsRDF, abUser);
            } // have urn
          } // for all skin content packages
        } // have skin content packages
      } // for all skins
    } // have skin packages

    // save all files in cache to disk
    for (i in oOverlayCache)
    {
      oOverlayCache[i].FlushDataSource(oOverlayCache[i].msDataSourceURI);
      delete oOverlayCache[i];
    }
  } // have chrome.rdf
  debugdump("UpdateAllOverlayInfo-");
}


// inner helper function
function UpdateOverlayInfoFile(asInfoFileURI, aoOverlayCache, aoContentsRDF, aoSeq)
{
  if (asInfoFileURI in aoOverlayCache)
  { // get RDF file
    var oInfoRDF = aoOverlayCache[asInfoFileURI];
    // create target container
    const koInnerSeq = aoContentsRDF.InitContainer(aoSeq);
    debugdump("  UpdateOverlayInfoFile: koInnerSeq="+koInnerSeq.Resource.Value);
    const koSourceSeq = koInnerSeq.Resource;
    var oTargetSeq = oInfoRDF.InitContainer(koSourceSeq);
    if (!oTargetSeq)
    { // create new Seq
      oInfoRDF.AddContainer(koSourceSeq);
      oTargetSeq = oInfoRDF.InitContainer(koSourceSeq);
    }
    if (oTargetSeq)
    { // have target <Seq>
      const koInnerEnum = koInnerSeq.GetElements();
      while (koInnerEnum.hasMoreElements())
      { // get source container
        const koLit = koInnerEnum.getNext();
        if (koLit instanceof goMnenhy.rdf.gkIRDFLiteral)
        {
          debugdump("   UpdateOverlayInfoFile: koLit="+koLit.Value);
          // set target literal
          if (oTargetSeq.IndexOf(koLit) == -1)
            oTargetSeq.AppendElement(koLit);
        } // have literal
      } // loop over overlayinfo entries
    } // have target <Seq>
  } // have overlayinfo file
}
  


// Update the overlayinfo with the information of the overlays and
// stylesheets sections of the contents.rdf given by aoContentsURI.
// Also update the new flattened overlayinfo files (bug 278534).
function UpdateOverlayInfo(aoOverlayCache, aoContentsURI, abUser)
{ // mark this overlayinfo recreation
  debugdump("UpdateOverlayInfo("+aoContentsURI+")+");
  if (aoOverlayCache && aoContentsURI)
  { // init the contents.rdf file to be read
    if (typeof(aoContentsURI) == "string")
      aoContentsURI = goMnenhy.m_koIOService.newURI(aoContentsURI, null, null);
    const koContentsRDF = new goMnenhy.rdf.MnenhyRDF(aoContentsURI, null, null, null, true);
    if (koContentsRDF.IsGood())
    { // loop over overlays and stylesheets
      const krsRootSeqs = [gksMozillaOverlaysRoot, gksMozillaStylesheetsRoot];
      for (var sRootSeq in krsRootSeqs)
      { // process the root <Seq> and write its data into the respective overlayinfo file
        const koOuterSeq = koContentsRDF.InitContainer(krsRootSeqs[sRootSeq]);
        if (koOuterSeq)
        { // root <Seq> found
          debugdump(" UpdateOverlayInfo[5]: RootSeq="+koOuterSeq.Resource.Value);
          const kbIsOverlay = (sRootSeq == "0");
          const koOuterEnum = koOuterSeq.GetElements();
          while (koOuterEnum.hasMoreElements())
          { // get source container
            const koSeq = koOuterEnum.getNext().QueryInterface(goMnenhy.rdf.gkIRDFResource);
            debugdump("  UpdateOverlayInfo[6]: koSeq="+koSeq);
            // get handle to target overlayinfo file
            var oInfoRDF = null;
            const ksInfoFileURI = GetOverlayInfoFile(koContentsRDF.msDataSourceURI, koSeq.Value, kbIsOverlay);
            debugdump("  UpdateOverlayInfo[7]: ksInfoFileURI="+ksInfoFileURI);
            if (!(ksInfoFileURI in aoOverlayCache))
            { // open new overlayinfo file
              oInfoRDF = new goMnenhy.rdf.MnenhyRDF(ksInfoFileURI, null, null, null, true);
              if (oInfoRDF)
              { // remember the RDF file
                aoOverlayCache[ksInfoFileURI] = oInfoRDF;
                // if we haven't touched it yet, empty it
                debugdump("  UpdateOverlayInfo[8]: emptying ksInfoFileURI");
                const koAllRes = oInfoRDF.moDataSource.GetAllResources();
                while (koAllRes.hasMoreElements())
                { // inspect every resource
                  const koRes = koAllRes.getNext();
                  oInfoRDF.RemoveNode(koRes);
                } // while all resources
              }
            }
            UpdateOverlayInfoFile(ksInfoFileURI, aoOverlayCache, koContentsRDF, koSeq);
            // now write into the flat file also
            const ksInfoFlatFileURI = (abUser ? goMnenhy.m_ksURIUser : goMnenhy.m_ksURIApp)
                                    + gksMozillaChrome
                                    + (kbIsOverlay ? gksMozillaOverlaysRDF : gksMozillaStylesheetsRDF);
            debugdump("  UpdateOverlayInfo[9]: ksInfoFlatFileURI="+ksInfoFlatFileURI);
            if (!(ksInfoFlatFileURI in aoOverlayCache))
            { // open new overlayinfo file
              oInfoRDF = new goMnenhy.rdf.MnenhyRDF(ksInfoFlatFileURI, null, null, null, true);
              if (oInfoRDF)
              { // remember the RDF file
                aoOverlayCache[ksInfoFlatFileURI] = oInfoRDF;
                // do not empty it!
              }
            }
            UpdateOverlayInfoFile(ksInfoFlatFileURI, aoOverlayCache, koContentsRDF, koSeq);
          } // loop over all overlays/stylesheets
        } // root <Seq> found
      } // for overlays/stylesheets
    } // have valid aoContentsURI
  }
  debugdump("UpdateOverlayInfo-");
}


// |
// |  helper functions
// |  (actually not really related to uninstall, but needed ;-) )
// |

// read all the prefs needed in this file
function ReadPrefs()
{
  debugdump("ReadPrefs+");
  gbNoBackup       = goMnenhy.GetPref(gksUninstallNoBackupPref,   goMnenhy.kiPrefBool);
  debugdump("ReadPrefs: gbNoBackup      ="+gbNoBackup);
  gbNoShutdown     = goMnenhy.GetPref(gksUninstallNoShutdownPref, goMnenhy.kiPrefBool);
  debugdump("ReadPrefs: gbNoShutdown    ="+gbNoShutdown);
  gbUninstallDebug = goMnenhy.GetPref(gksUninstallDebugPref,      goMnenhy.kiPrefBool);
  debugdump("ReadPrefs: gbUninstallDebug="+gbUninstallDebug);
  debugdump("ReadPrefs-");
}


// Do we have administra(i)tor rights for the Mozilla install chrome directory?
//
// return value:
//    true:     we are root!
//    false:    we're in user hell.
function IsGodMode()
{ // try to write a file into the install chrome directory and delete it
  debugdump("IsGodMode+");
  var bGod = false;
  try
  { // create file
    const klGodPermissions = 0644;  // only owner may change, others only read
    var oGodFile = goMnenhy.m_koChromeDirApp.clone();
    oGodFile.append(gksGodMode)
    oGodFile.create(oGodFile.NORMAL_FILE_TYPE, klGodPermissions);
    // delete the file
    oGodFile.remove(false);
    bGod = true;
  }
  catch(ignored)
  {
    debugdump(ignored);
  }
  debugdump("IsGodMode-: "+bGod);
  return bGod;
}


// get the chrome.rdf
// if aoPath is a non-null nsIFile, look there
// else try the user profile dir (abUser = true) or the app chrome dir (abUser = false).
function GetChromeRDF(abUser, aoPath)
{
  debugdump("GetChromeRDF+");
  var oRDF  = null;
  // get the path
  var oPath = aoPath;
  if (!oPath)
    oPath = abUser ? goMnenhy.m_koChromeDirUser : goMnenhy.m_koChromeDirApp;
  // get the chrome.rdf file
  var oChrome = oPath.clone();
  if (oChrome)
  { // open it as RDF
    oChrome.append(gksMozillaChromeRDF);
    if (oChrome.exists())
    {
      oChrome = goMnenhy.GetProtocolHandler().newFileURI(oChrome);
      oRDF = new goMnenhy.rdf.MnenhyRDF(oChrome, null, gksMozillaNamespace, gkrsMozillaBaseCont);
    }
  }
  debugdump("GetChromeRDF-");
  return oRDF;
}


// create a line for use in installed-chrome.txt
// uri must not be empty; defaults for the other parametres are content, install, url, LF
// asLocation accepts |true| as an alias for gksLocProfile
function CreateICLine(asURI, asProvider, asLocation, asURIType, asLF)
{ // very thin ice indeed ;-)
  debugdump("CreateICLine("+asURI+", "+asProvider+", "+asLocation+", "+asURIType+", "+asLF+")+");
  var sLine = null;
  if (asURI)
  {
    // default provider is content
    switch (asProvider)
    {
      case gksTypeContent:
      case gksTypeSkin:
      case gksTypeLocale:
        break;
      default:
        asProvider = gksTypeContent;
    }
    // default location is application
    switch (asLocation)
    {
      case gksLocInstall:
      case gksLocProfile:
      case gksLocSelect:
        break;
      case true:
        asLocation = gksLocProfile;
        break;
      default:
        asLocation = gksLocInstall;
    }
    // default URI type is url
    switch (asURIType)
    {
      case gksURIURL:
      case gksURIPath:
        break;
      default:
        asURIType = gksURIURL;
    }
    // default LF is LF
    if (!asLF)
      asLF = gksLF
    // create
    sLine = asProvider + "," + asLocation + "," + asURIType + "," + asURI + asLF;
  }
  debugdump("CreateICLine-");
  return sLine;
}


// get the installed-chrome.txt as a nsIFile
function GetInstalledChromeFile()
{
  debugdump("GetInstalledChromeFile+");
  var oFile = goMnenhy.m_koChromeDirApp.clone();
  oFile.append(gksMozillaInstChromeTxt);
  debugdump("GetInstalledChromeFile-");
  return oFile;
}
// get the contents of installed-chrome.txt
function LoadInstalledChrome()
{
  debugdump("LoadInstalledChrome");
  return goMnenhy.ReadFile(GetInstalledChromeFile());
}
// save sContent to installed-chrome.txt
function SaveInstalledChrome(sContent)
{ // clean-up: remove all CR, remove whitespace lines, keep only one LF, expand LF to CRLF
  debugdump("SaveInstalledChrome+");
  if (!sContent)
  {
    sContent = "";
  }
  else
  {
    sContent = sContent.replace(/\r/g,    "");
    sContent = sContent.replace(/\s*\n/g, gksLF);
    sContent = sContent.replace(/(\n)+/g, gksLF);
    sContent = sContent.replace(/\n/g,    gksCRLF);
    sContent += gksCRLF;
  }
  // save
  goMnenhy.WriteFile(GetInstalledChromeFile(), sContent);
  debugdump("SaveInstalledChrome-");
}


// Create the overlayinfo file URI string for either overlays or stylesheets
// IN:     string or nsIURI  aoSourceURI  the contents.rdf with the list of files to be overlayed
// IN:     string or nsIURI  aoTargetURI  the file to be overlayed
// IN:     bool              abOverlays   whether the file is an overlay (XUL) or stylesheet (CSS)
// RETURN: string            sFileURI     file URI of the respective overlayinfo file
function GetOverlayInfoFile(aoSourceURI, aoTargetURI, abOverlays)
{
  debugdump("GetOverlayInfoFile("+aoSourceURI+","+aoTargetURI+","+abOverlays+")+");
  var sFileURI = "";
  if (aoSourceURI && aoTargetURI)
  {
    if (typeof(aoTargetURI) == "string")
      aoTargetURI = goMnenhy.m_koIOService.newURI(aoTargetURI, null, null);

    const ksSrcFileURI = goMnenhy.ConvertURI2FileURI(aoSourceURI);
    if (ksSrcFileURI && aoTargetURI && aoTargetURI.schemeIs("chrome"))
    { // canonify the chrome uri
      // interface change around ~2005-02-25!
      if ("canonify" in goMnenhy.m_koChromeRegService)
        goMnenhy.m_koChromeRegService.canonify(aoTargetURI);
      else
        aoTargetURI = goMnenhy.GetProtocolHandler(goMnenhy.ksProtocolChrome).newURI(aoTargetURI.spec, null, null);
      // get the package name
      const krPack = aoTargetURI.spec.match("^chrome://(.*?)(?=\.jar|/)");
      if (1 in krPack)
      { // found package name
        const koSrcInfo = ChromeURI2Info(ksSrcFileURI.spec);
        sFileURI = (koSrcInfo.bProfile ? goMnenhy.m_ksURIUser : goMnenhy.m_ksURIApp)
                 + gksMozillaChrome
                 + gksMozillaOverlayInfo
                 + krPack[1]
                 + (abOverlays ? gksMozillaOverlaysRDFPath : gksMozillaStylesheetsRDFPath);
      } // found package name
    } // have ksSrcFileURI and chrome target
  }
  debugdump("GetOverlayInfoFile["+sFileURI+"]-");
  return sFileURI;
}


// create a nsIFile from the given aoURI (given as string or nsIURI)
function ConvertURI2File(aoURI)
{
  debugdump("ConvertURI2File+");
  var oFile = null;
  if (aoURI)
  {
    const koFileURI = goMnenhy.ConvertURI2FileURI(aoURI);
    if (koFileURI)
    { // create nsIFile from URI
      oFile = koFileURI.file.clone();
    }
  }
  debugdump("ConvertURI2File-");
  return oFile;
}


// remove $UserProfileDir/xul.mfl (resp. xul.mfasl)
function DeleteXulFastload()
{
  debugdump("DeleteXulFastload+");
  try
  {
    var oFaslService = Components.classes[gksCID_Fastload]
                                 .getService(Components.interfaces.nsIFastLoadService);
    var oFaslFile    = oFaslService.newFastLoadFile(gksMozillaXUL_Fastload);
    oFaslFile.remove(false);
  }
  catch(ignored){}
  debugdump("DeleteXulFastload-");
}


// SeaMonkey has a dedicated function for temporarily deactivating certain packages -
// this has been dropped in the new toolkit registry. We simulate this by keeping that package in
// the manifest file, but commenting it out.
function SetAllowOverlaysForPackage(asPackage, abAllowed)
{
  debugdump("uninstall.SetAllowOverlaysForPackage(" + asPackage + "," + abAllowed +")");
  if (goMnenhy.m_kbIsToolkitChromeRegistry)
  { // comment this package's lines out
    const koMnenhyManifestFile = GetMnenhyManifestFile();
    if (koMnenhyManifestFile)
    {
      var sMnenhyManifest = "\n" + goMnenhy.ReadFile(koMnenhyManifestFile) + "\n";
      var bWriteManifest  = false;
      // does the manifest file contain our package?
      const ksPackageLine = "\\n\\s*#?\\s*((content|locale|skin)\\s+" + asPackage + "\\s.*)";
      const kxPackageLine = new RegExp(ksPackageLine);
      // the global search gets all matching lines, but no brace-matches
      const krsPackageLines = sMnenhyManifest.match(new RegExp(ksPackageLine, "g"));
      debugdump("uninstall.SetAllowOverlaysForPackage: krsPackageLines=" + krsPackageLines);
      for (var i in krsPackageLines)
      { // (don't) comment these lines out (if match[1] wouldn't exist, we weren't here in the first place...)
        const ksNewPackageLine = "\n" + (abAllowed ? "" : "#") + krsPackageLines[i].match(kxPackageLine)[1];
        sMnenhyManifest = sMnenhyManifest.replace(krsPackageLines[i], ksNewPackageLine);
      }
      // we need to touch all overlays and overrides also
      const ksOverLine = "\\n\\s*#?\\s*((overlay|override)\\s+.*?\\s+chrome://" + asPackage + "/.*)";
      const kxOverLine = new RegExp(ksOverLine);
      // the global search gets all matching lines, but no brace-matches
      const krsOverLines = sMnenhyManifest.match(new RegExp(ksOverLine, "g"));
      debugdump("uninstall.SetAllowOverlaysForPackage: krsOverKnown=" + krsOverLines);
      for (var j in krsOverLines)
      { // (don't) comment these lines out (if match[1] wouldn't exist, we weren't here in the first place...)
        const ksNewOverLine = "\n" + (abAllowed ? "" : "#") + krsOverLines[j].match(kxOverLine)[1];
        sMnenhyManifest = sMnenhyManifest.replace(krsOverLines[j], ksNewOverLine);
      }
      // and write the manifest back
      if (krsPackageLines || krsOverLines)
      {
        try
        { // remove enclosing whitespace
          sMnenhyManifest = sMnenhyManifest.replace(/^\s+|\s+$/g, "");
          goMnenhy.WriteFile(koMnenhyManifestFile, sMnenhyManifest);
        }
        catch (ignored) 
        { // couldn't change the manifest, most probably because we lack write permissions
          debugdump("uninstall.SetAllowOverlaysForPackage: couldn't update Mnenhy's chrome.manifest!\n" +
                    "Insufficient write permissions for '" + koMnenhyManifestFile.spec + "'?");
        }
      } // if bWriteManifest
    }
  }
  else
  { // Mozilla/Seamonkey know how to handle that themselves
    goMnenhy.m_koChromeRegService.setAllowOverlaysForPackage(asPackage, abAllowed);
  }
}


// Get the toolkit chrome.manifest file for Mnenhy
// Return: an nsIFile
function GetMnenhyManifestFile()
{
  var oRet = null;
  if (goMnenhy.m_kbIsToolkitChromeRegistry)
  { // chrome.rdf is no more, so interprete Mnenhy's chrome.manifest:
    // - installed packages appear there as a "content" line
    // - inactive packages appear there as comment-perefixed (#) "content" line
    const krsMnenhyManifest = goMnenhy.ConvertURI2FileURI("chrome://mnenhy/content/").spec.match(/(.*\/chrome)\/.*$/);
    debugdump("uninstall.GetMnenhyManifestFile: krsMnenhyManifest="+krsMnenhyManifest);
    if (krsMnenhyManifest && krsMnenhyManifest.length > 1)
    { // \1 should be something like "file:///path/to/Mnenhy-GUID/chrome"
      oRet = goMnenhy.ConvertURI2FileURI(krsMnenhyManifest[1] + ".manifest").file;
    }
  }
  debugdump("uninstall.GetMnenhyManifestFile="+oRet);
  return oRet;
}


// debug
function debugdump(aoDump)
{
  if (gbUninstallDebug)
    goMnenhy.Log(aoDump);
}


// |
// |  Init called by the global Mnenhy object when loading this script file
// |
function Init(aoMnenhy)
{ // keep pointer to have the Mnenhy object prototype available
  this.goMnenhy     = aoMnenhy;
  // get certain flags
  ReadPrefs();
  // allow app chrome modifications?
  this.mbIsGodMode  = IsGodMode();
  // chrome RDF files
  this.mroChromeRDF = [GetChromeRDF(true), (mbIsGodMode ? GetChromeRDF(false) : null)];
  // get the shutdown function into the hidden window
  // make it act when it's safe, ie after a timeout
  goMnenhy.Include("chrome://global/content/globalOverlay.js", this);
  this.goQuitApplicationTimed = this.goQuitApplication;
  this.goQuitApplication = function(){setTimeout(this.goQuitApplicationTimed, 10);}
}